'use strict';
var path = require('path');
var events = require('events');
var anyMock = require('ruff-mock').anyMock;

var include = function(target, mixInObj) {
    var mockObject = target;
    for (var fm in mixInObj) {
        if (mixInObj[fm] === 'constructor') {
            continue;
        } else {
            mockObject[fm] = mixInObj[fm];
        }
    }
    return mockObject;
};

var mockMissing = function(obj) {
    return include(anyMock(), obj);
};

function safeRequire(filename) {
    try {
        return require(filename);
    } catch (error) {
        if (error instanceof SyntaxError) {
            throw new SyntaxError('Fail to parse ' + filename + ', ' + error.message);
        }

        throw error;
    }
}

function AppRunner() {
    this._devices = {};
    this._currentAppPath = null;

    this._appEnd = null;
    this._appReady = null;
}

AppRunner.prototype._getDevice = function(id) {
    if (!(id in this._devices)) {
        throw Error('No [' + id + '] found in app.json');
    }
    return this._devices[id];
};

AppRunner.prototype._setAppEnd = function(cb) {
    this._appEnd = cb;
};

AppRunner.prototype._setAppReady = function(cb) {
    this._appReady = cb;
};

function mock$(appRunner) {
    function $(devId) {
        return appRunner._getDevice(devId);
    }

    $.ready = function(cb) {
        appRunner._setAppReady(cb);
    };

    $.end = function(cb) {
        appRunner._setAppEnd(cb);
    };

    return $;
}

AppRunner.prototype._addDevice = function(id) {
    this._devices[id] = mockMissing(new events.EventEmitter());
};

AppRunner.prototype._loadBoard = function(board) {
    var boardConfig = safeRequire(path.join(this._currentAppPath, 'ruff_modules', board, 'board.json'));

    if (boardConfig && boardConfig.devices) {
        var self = this;
        boardConfig.devices.forEach(function(device) {
            self._addDevice(device.id);
        });
        return;
    }

    throw new Error('Fail to parse board, please ensure board has been installed correctly');
};

AppRunner.prototype._loadConfig = function() {
    var self = this;
    var appConfig = safeRequire(path.join(this._currentAppPath, 'app.json'));
    if (appConfig && appConfig.board && appConfig.devices) {
        this._loadBoard(appConfig.board);
        appConfig.devices.forEach(function(device) {
            self._addDevice(device.id);
        });

        return;
    }

    throw new Error('Fail to parse app.json, please ensure project has been configured correctly');
};

AppRunner.prototype._doRun = function(appPath, runnable, args) {
    var container = this;
    this._currentAppPath = appPath;
    this._devices = {};

    this._loadConfig();

    global.$ = mock$(this);

    safeRequire(path.join(appPath, 'src', 'index.js'));

    this._appReady();
    runnable.apply(container, args);
    this._appEnd();

    return {
        end : function(endCallback) {
            endCallback();
        }
    };
};

function runnable(one) {
    if (typeof one === 'function') {
        return one;
    }

    throw new TypeError('Runnable function is expected to run app runner');
}

AppRunner.prototype.run = function() {
    var appPath = arguments[0];
    var targetRunnable = runnable(arguments[1]);
    var leftArgIndex = 2;

    return this._doRun(appPath, targetRunnable,
        Array.prototype.slice.call(arguments, leftArgIndex));
};

module.exports = new AppRunner();