/*jshint bitwise: false*/
/*jshint -W003 */
'use strict';

var Stream        = require('stream');
var Buffer        = require('buffer').Buffer;
var util          = require('util');
var debug         = function () {
};
var StringDecoder = require('string_decoder.js').StringDecoder;

// set options on this.
function ReadableState(options, stream) {
	options = options || {};

	// object stream flag. Used to make read(n) ignore n and to
	// make all the buffer merging and length checks go away
	this.objectMode = !!options.objectMode;

	if (stream instanceof Stream.Duplex) {
		this.objectMode = this.objectMode || !!options.readableObjectMode;
	}

	// the point at which it stops calling _read() to fill the buffer
	// Note: 0 is a valid value, means "don't call _read preemptively ever"
	var hwm            = options.highWaterMark;
	var defaultHwm     = this.objectMode ? 16 : 16 * 1024;
	this.highWaterMark = (hwm || hwm === 0) ? hwm : defaultHwm;

	// cast to int.
	this.highWaterMark = ~~this.highWaterMark;

	this.buffer     = [];
	this.length     = 0;
	this.flowing    = null;
	this.ended      = false;
	this.endEmitted = false;
	this.reading    = false;

	// a flag to be able to tell if the onwrite cb is called immediately,
	// or on a later tick.  We set this to true at first, because any
	// actions that shouldn't happen until "later" should generally also
	// not happen before the first write call.
	this.sync = true;

	// whenever we return null, then we set a flag to say
	// that we're awaiting a 'readable' event emission.
	this.needReadable      = false;
	this.emittedReadable   = false;
	this.readableListening = false;

	// Crypto is kind of old and crusty.  Historically, its default string
	// encoding is 'binary' so we have to make this configurable.
	// Everything else in the universe uses 'utf8', though.
	this.defaultEncoding = options.defaultEncoding || 'utf8';

	// if true, a maybeReadMore has been scheduled
	this.readingMore = false;

	this.decoder  = null;
	this.encoding = null;
	if (options.encoding) {
		this.decoder  = new StringDecoder(options.encoding);
		this.encoding = options.encoding;
	}
}

function Readable(options) {
	if (!(this instanceof Readable)) {
		return new Readable(options);
	}

	this._readableState = new ReadableState(options, this);

	// legacy
	this.readable = true;

	if (options && typeof options.read === 'function') {
		this._read = options.read;
	}

	Stream.call(this);
}

util.inherits(Readable, Stream);

// if it's past the high water mark, we can push in some more.
// Also, if we have no data yet, we can stand some
// more bytes.  This is to work around cases where hwm=0,
// such as the repl.  Also, if the push() triggered a
// readable event, and the user called read(largeNumber) such that
// needReadable was set, then we ought to push more, so that another
// 'readable' event will be triggered.
function needMoreData(state) {
	return !state.ended &&
		(state.needReadable ||
		state.length < state.highWaterMark ||
		state.length === 0);
}

function isChunkInvalid(chunk, state) {
	return !(chunk instanceof Buffer) &&
		typeof chunk !== 'string' &&
		chunk !== null &&
		chunk !== undefined && !state.objectMode;
}

function checkChunk(stream, state, chunk) {
	if (isChunkInvalid(chunk, state)) {
		stream.emit('error', new TypeError('Invalid non-string/buffer chunk'));
		return false;
	}
	return true;
}

function readChunkInObjectMode(state, addToFront, stream, encoding, chunk) {
	var e;
	if (state.ended && !addToFront) {
		e = new Error('stream.push() after EOF');
		stream.emit('error', e);
	} else if (state.endEmitted && addToFront) {
		e = new Error('stream.unshift() after end event');
		stream.emit('error', e);
	} else {
		if (state.decoder && !addToFront && !encoding) {
			chunk = state.decoder.write(chunk);
		}

		if (!addToFront) {
			state.reading = false;
		}

		// if we want the data now, just emit it.
		if (state.flowing && state.length === 0 && !state.sync) {
			stream.emit('data', chunk);
			stream.read(0);
		} else {
			// update the buffer info.
			state.length += state.objectMode ? 1 : chunk.length;
			if (addToFront) {
				state.buffer.unshift(chunk);
			}
			else {
				state.buffer.push(chunk);
			}

			if (state.needReadable) {
				emitReadable(stream);
			}
		}

		maybeReadMore(stream, state);
	}
	return chunk;
}
function canReadChunkInObjectMode(state, chunk) {
	return state.objectMode || chunk && chunk.length > 0;
}
function endReading(state, stream) {
	state.reading = false;
	onEofChunk(stream, state);
}
function readableAddChunk(stream, state, chunk, encoding, addToFront) {

	if (checkChunk(stream, state, chunk)) {

		if (chunk === null) {
			endReading(state, stream);
		} else if (canReadChunkInObjectMode(state, chunk)) {
			readChunkInObjectMode(state, addToFront, stream, encoding, chunk);
		} else if (!addToFront) {
			state.reading = false;
		}
	}

	return needMoreData(state);
}

// Manually shove something into the read() buffer.
// This returns true if the highWaterMark has not been hit yet,
// similar to how Writable.write() returns true if you should
// write() some more.
Readable.prototype.push = function (chunk, encoding) {
	var state = this._readableState;

	if (!state.objectMode && typeof chunk === 'string') {
		encoding = encoding || state.defaultEncoding;
		if (encoding !== state.encoding) {
			chunk    = new Buffer(chunk, encoding);
			encoding = '';
		}
	}

	return readableAddChunk(this, state, chunk, encoding, false);
};

// Unshift should *always* be something directly out of read()
Readable.prototype.unshift = function (chunk) {
	var state = this._readableState;
	return readableAddChunk(this, state, chunk, '', true);
};

// Don't raise the hwm > 8MB
var MAX_HWM = 0x800000;
function computeNewHighWaterMark(n) {
	if (n >= MAX_HWM) {
		n = MAX_HWM;
	} else {
		// Get the next highest power of 2
		n--;
		n |= n >>> 1;
		n |= n >>> 2;
		n |= n >>> 4;
		n |= n >>> 8;
		n |= n >>> 16;
		n++;
	}
	return n;
}

function howMuchToRead(n, state) {
	if (state.length === 0 && state.ended) {
		return 0;
	}

	if (state.objectMode) {
		return n === 0 ? 0 : 1;
	}

	if (n === null || isNaN(n)) {
		// only flow one buffer at a time
		if (state.flowing && state.buffer.length) {
			return state.buffer[0].length;
		}
		else {
			return state.length;
		}
	}

	if (n <= 0) {
		return 0;
	}

	// If we're asking for more than the target buffer level,
	// then raise the water mark.  Bump up to the next highest
	// power of 2, to prevent increasing it excessively in tiny
	// amounts.
	if (n > state.highWaterMark) {
		state.highWaterMark = computeNewHighWaterMark(n);
	}

	// don't have that much.  return null, unless we've ended.
	if (n > state.length) {
		if (!state.ended) {
			state.needReadable = true;
			return 0;
		} else {
			return state.length;
		}
	}

	return n;
}

// you can override either this method, or the async _read(n) below.
Readable.prototype.read = function (n) {
	debug('read', n);

	var state = this._readableState;

	var nOrig = n;

	if (typeof n !== 'number' || n > 0) {
		state.emittedReadable = false;
	}

	// if we're doing read(0) to trigger a readable event, but we
	// already have a bunch of data in the buffer, then just trigger
	// the 'readable' event and move on.
	if (n === 0 &&
		state.needReadable &&
		(state.length >= state.highWaterMark || state.ended)) {
		debug('read: emitReadable', state.length, state.ended);
		if (state.length === 0 && state.ended) {
			endReadable(this);
		}
		else {
			emitReadable(this);
		}
		return null;
	}

	n = howMuchToRead(n, state);

	// if we've ended, and we're now clear, then finish it up.
	if (n === 0 && state.ended) {
		if (state.length === 0) {
			endReadable(this);
		}
		return null;
	}

	// All the actual chunk generation logic needs to be
	// *below* the call to _read.  The reason is that in certain
	// synthetic stream cases, such as passthrough streams, _read
	// may be a completely synchronous operation which may change
	// the state of the read buffer, providing enough data when
	// before there was *not* enough.
	//
	// So, the steps are:
	// 1. Figure out what the state of things will be after we do
	// a read from the buffer.
	//
	// 2. If that resulting state will trigger a _read, then call _read.
	// Note that this may be asynchronous, or synchronous.  Yes, it is
	// deeply ugly to write APIs this way, but that still doesn't mean
	// that the Readable class should behave improperly, as streams are
	// designed to be sync/async agnostic.
	// Take note if the _read call is sync or async (ie, if the read call
	// has returned yet), so that we know whether or not it's safe to emit
	// 'readable' etc.
	//
	// 3. Actually pull the requested chunks out of the buffer and return.

	// if we need a readable event, then we need to do some reading.
	var doRead = state.needReadable;
	debug('need readable', doRead);

	// if we currently have less than the highWaterMark, then also read some
	if (state.length === 0 || state.length - n < state.highWaterMark) {
		doRead = true;
		debug('length less than watermark', doRead);
	}

	// however, if we've ended, then there's no point, and if we're already
	// reading, then it's unnecessary.
	if (state.ended || state.reading) {
		doRead = false;
		debug('reading or ended', doRead);
	}

	if (doRead) {
		debug('do read');
		state.reading = true;
		state.sync    = true;
		// if the length is currently zero, then we *need* a readable event.
		if (state.length === 0) {
			state.needReadable = true;
		}
		// call internal read method
		this._read(state.highWaterMark);
		state.sync    = false;
	}

	// If _read pushed data synchronously, then `reading` will be false,
	// and we need to re-evaluate how much data we can return to the user.
	if (doRead && !state.reading) {
		n = howMuchToRead(nOrig, state);
	}

	var ret;
	if (n > 0) {
		ret = fromList(n, state);
	}
	else {
		ret = null;
	}

	if (ret === null) {
		state.needReadable = true;
		n                  = 0;
	}

	state.length -= n;

	// If we have nothing in the buffer, then we want to know
	// as soon as we *do* get something into the buffer.
	if (state.length === 0 && !state.ended) {
		state.needReadable = true;
	}

	// If we tried to read() past the EOF, then emit end on the next tick.
	if (nOrig !== n && state.ended && state.length === 0) {
		endReadable(this);
	}

	if (ret !== null) {
		this.emit('data', ret);
	}

	return ret;
};

function onEofChunk(stream, state) {
	if (state.ended) {
		return;
	}
	if (state.decoder) {
		var chunk = state.decoder.end();
		if (chunk && chunk.length) {
			state.buffer.push(chunk);
			state.length += state.objectMode ? 1 : chunk.length;
		}
	}
	state.ended = true;

	// emit 'readable' now to make sure it gets picked up.
	emitReadable(stream);
}

// Don't emit readable right away in sync mode, because this can trigger
// another read() call => stack overflow.  This way, it might trigger
// a nextTick recursion warning, but that's not so bad.
function emitReadable(stream) {
	var state          = stream._readableState;
	state.needReadable = false;
	if (!state.emittedReadable) {
		debug('emitReadable', state.flowing);
		state.emittedReadable = true;
		if (state.sync) {
			process.nextTick(emitReadable_, stream);
		} else {
			emitReadable_(stream);
		}
	}
}

function emitReadable_(stream) {
	debug('emit readable');
	stream.emit('readable');
	flow(stream);
}

// at this point, the user has presumably seen the 'readable' event,
// and called read() to consume some data.  that may have triggered
// in turn another _read(n) call, in which case reading = true if
// it's in progress.
// However, if we're not ended, or reading, and the length < hwm,
// then go ahead and try to read some more preemptively.
function maybeReadMore(stream, state) {
	if (!state.readingMore) {
		state.readingMore = true;
		process.nextTick(maybeReadMore_, stream, state);
	}
}

function maybeReadMore_(stream, state) {
	var len = state.length;
	while (!state.reading && !state.flowing && !state.ended &&
	state.length < state.highWaterMark) {
		debug('maybeReadMore read 0');
		// rere
		stream.read(0);
		if (len === state.length)
		// didn't get any data, stop spinning.
		{
			break;
		} else {
			len = state.length;
		}
	}
	state.readingMore = false;
}

// abstract method.  to be overridden in specific implementation classes.
// call cb(er, data) where data is <= n in length.
// for virtual (non-string, non-buffer) streams, "length" is somewhat
// arbitrary, and perhaps not very meaningful.
Readable.prototype._read = function (n) {// jshint ignore:line
	this.emit('error', new Error('not implemented'));
};

// set up data events if they are asked for
// Ensure readable listeners eventually get something
Readable.prototype.on          = function (ev, fn) {
	var res = Stream.prototype.on.call(this, ev, fn);

	// If listening to data, and it has not explicitly been paused,
	// then call resume to start the flow of data on the next tick.
	if (ev === 'data' && false !== this._readableState.flowing) {
		this.resume();
	}

	if (ev === 'readable' && this.readable) {
		var state = this._readableState;
		if (!state.readableListening) {
			state.readableListening = true;
			state.emittedReadable   = false;
			state.needReadable      = true;
			if (!state.reading) {
				process.nextTick(nReadingNextTick, this);
				//timers.setTimeout(nReadingNextTick);
			} else if (state.length) {
				emitReadable(this, state);
			}
		}
	}

	return res;
};
Readable.prototype.addListener = Readable.prototype.on;

function nReadingNextTick(self) {
	debug('readable nexttick read 0');
	self.read(0);
}

// pause() and resume() are remnants of the legacy readable stream API
// If the user uses them, then switch into old mode.
Readable.prototype.resume = function () {
	var state = this._readableState;
	if (!state.flowing) {
		debug('resume');
		state.flowing = true;
		resume(this, state);
	}
	return this;
};

function resume(stream, state) {
	if (!state.resumeScheduled) {
		state.resumeScheduled = true;
		process.nextTick(resume_, stream, state);
	}
}

function resume_(stream, state) {
	if (!state.reading) {
		debug('resume read 0');
		stream.read(0);
	}

	state.resumeScheduled = false;
	stream.emit('resume');
	flow(stream);
	if (state.flowing && !state.reading) {
		stream.read(0);
	}
}

Readable.prototype.pause = function () {
	debug('call pause flowing=%j', this._readableState.flowing);
	if (false !== this._readableState.flowing) {
		debug('pause');
		this._readableState.flowing = false;
		this.emit('pause');
	}
	return this;
};

function flow(stream) {
	var state = stream._readableState;
	debug('flow', state.flowing);
	if (state.flowing) {
		var chunk;
		do {
			chunk = stream.read();
		} while (null !== chunk && state.flowing);
	}
}

// wrap an old-style stream as the async data source.
// This is *not* part of the readable stream interface.
// It is an ugly unfortunate mess of history.
Readable.prototype.wrap = function (stream) {
	var state  = this._readableState;
	var paused = false;

	var self = this;
	stream.on('end', function () {
		debug('wrapped end');
		if (state.decoder && !state.ended) {
			var chunk = state.decoder.end();
			if (chunk && chunk.length) {
				self.push(chunk);
			}
		}

		self.push(null);
	});

	stream.on('data', function (chunk) {
		debug('wrapped data');
		if (state.decoder) {
			chunk = state.decoder.write(chunk);
		}

		// don't skip over falsy values in objectMode
		if (state.objectMode && (chunk === null || chunk === undefined)) {
			return;
		} else if (!state.objectMode && (!chunk || !chunk.length)) {
			return;
		}

		var ret = self.push(chunk);
		if (!ret) {
			paused = true;
			stream.pause();
		}
	});

	// proxy all the other methods.
	// important when wrapping filters and duplexes.
	/*jshint -W083 */
	/*jshint -W062 */
	for (var i in stream) {
		if (this[i] === undefined && typeof stream[i] === 'function') {
			this[i] = function (method) {
				return function () {
					return stream[method].apply(stream, arguments);
				};
			}(i);
		}
	}

	// proxy certain important events.
	var events = ['error', 'close', 'destroy', 'pause', 'resume'];
	events.forEach(function (ev) {
		stream.on(ev, self.emit.bind(self, ev));
	});

	// when we try to consume some more bytes, simply unpause the
	// underlying stream.
	self._read = function (n) {
		debug('wrapped _read', n);
		if (paused) {
			paused = false;
			stream.resume();
		}
	};

	return self;
};

// exposed for testing purposes only.
Readable._fromList = fromList;

// Pluck off n bytes from an array of buffers.
// Length is the combined lengths of all the buffers in the list.
function fromList(n, state) {
	var list       = state.buffer;
	var length     = state.length;
	var stringMode = !!state.decoder;
	var objectMode = !!state.objectMode;
	var ret;

	// nothing in the list, definitely empty.
	if (list.length === 0) {
		return null;
	}

	if (length === 0) {
		ret = null;
	} else if (objectMode) {
		ret = list.shift();
	} else if (!n || n >= length) {
		// read it all, truncate the array.
		if (stringMode) {
			ret = list.join('');
		} else {
			ret = Buffer.concat(list, length);
		}
		list.length = 0;
	} else {
		var buf;
		// read just some of it.
		if (n < list[0].length) {
			// just take a part of the first list item.
			// slice is the same for buffers and strings.
			buf     = list[0];
			ret     = buf.slice(0, n);
			list[0] = buf.slice(n);
		} else if (n === list[0].length) {
			// first list is a perfect match
			ret = list.shift();
		} else {
			// complex case.
			// we have enough to cover it, but it spans past the first buffer.
			if (stringMode) {
				ret = '';
			} else {
				ret = new Buffer(n);
			}

			var c = 0;
			for (var i = 0, l = list.length; i < l && c < n; i++) {
				buf     = list[0];
				var cpy = Math.min(n - c, buf.length);

				if (stringMode) {
					ret += buf.slice(0, cpy);
				} else {
					buf.copy(ret, c, 0, cpy);
				}

				if (cpy < buf.length) {
					list[0] = buf.slice(cpy);
				} else {
					list.shift();
				}

				c += cpy;
			}
		}
	}

	return ret;
}

function endReadable(stream) {
	var state = stream._readableState;

	// If we get here before consuming all the bytes, then that is a
	// bug in node.  Should never happen.
	if (state.length > 0) {
		throw new Error('endReadable called on non-empty stream');
	}

	if (!state.endEmitted) {
		state.ended = true;
		process.nextTick(endReadableNT, state, stream);
	}
}

function endReadableNT(state, stream) {
	// Check that we didn't get one last unshift.
	if (!state.endEmitted && state.length === 0) {
		state.endEmitted = true;
		stream.readable  = false;
		stream.emit('end');
	}
}

module.exports         = Readable;
Readable.ReadableState = ReadableState;
